#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Nov 21 16:23:15 2019

@author: mpolikarpov
"""
import os
os.environ['OMP_NUM_THREADS'] ='1'
os.environ['OPENBLAS_NUM_THREADS'] = '1'
os.environ['MKL_NUM_THREADS'] = '1'

from maximus48 import var
import numpy as np
from multiprocessing import Array
from maximus48.multiCTF2 import shift_distance as shift
 
"""
some functions here are out of classes - if you compare to the file tomo_proc.py
the reason for that - I wanted to make clases F and Processor as light as possible
 for parallel processing 
so, in a sence, I use them almost as fast dictionaries
"""

# =============================================================================
# tomo-functions
# =============================================================================

def tonumpyarray(shared_array, shape, dtype):
    '''Create numpy array from shared memory.'''
    nparray = np.frombuffer(shared_array, dtype=dtype).reshape(shape)    
    #assert nparray.base is shared_array
    return nparray
        

def correct_shifts(shifts, median_dev = 5):
    """find any bad numbers which are deviate more than 5 pixels from the median
    and correct them to median of the array"""
    
    shifts = tonumpyarray(shifts.shared_array_base, shifts.shape, shifts.dtype)
    
    for i in range(shifts.shape[1]):
        for j in range(shifts.shape[2]):
            shifts[:,i,j] = np.where((abs(shifts[:,i,j] - np.median(shifts[:,i,j])) > 5), np.median(shifts[:,i,j]), shifts[:,i,j])
    print('adjusted shifts') 


def rotaxis(proj, N_steps, ROI_ff = None):
    """calculate the rotation axis comparing 0 and 180 projection shift
    proj: 3D array
    N_steps: projections per degree
    N_rot: number of files that have their mirrors
    ROI_ff: where to make comparison
    
    """
    if not ROI_ff:
        ROI_ff = 0,0, proj.shape[1], proj.shape[0]
        
    cent = []
    N_rot = proj.shape[0] - 180 * N_steps
    
    for i in range(N_rot):
        distances = shift(proj[i, ROI_ff[1]:ROI_ff[3], ROI_ff[0]:ROI_ff[2]], 
              np.flip(proj[i + N_steps*180, ROI_ff[1]:ROI_ff[3], ROI_ff[0]:ROI_ff[2]] ,1))
        cent.append(proj[i].shape[1]/2 + distances[1]/2)
    return cent
    

        

# =============================================================================
# #actually should be a part of the Processor class - check tomo_proc.py
# =============================================================================

def init_Npad(ROI, compression = 8):
    """Calculate the Npad for padding
    can be adjusted with compression parameter
    By default, 8 times smaller than ROI
    """
                    
    if (ROI[2]-ROI[0])>(ROI[3]-ROI[1]):
        Npad = (ROI[2]-ROI[0])//compression       
    else:
        Npad = (ROI[3]-ROI[1])//compression   
        
    return Npad 


def init_names(data_name, N_distances, first_distance = 1):
    """set proper data_names"""
    
    data_names = []
    ff_names = []
    
    for i in range(first_distance, N_distances + first_distance):
        data_names.append(data_name + '_' + str(i))
        ff_names.append('ff_' + data_name + '_' + str(i))
        
    return data_names, ff_names 







# =============================================================================
# classes themselves
# =============================================================================
class Processor:
    __slots__ = ['ROI', 'ROI_ff', 'Npad', 'im_shape', 'images', 'flats',
                 'N_files', 'N_start']
        
    def __init__(self, ROI, folder, N_start, N_finish, compNpad = 8):
        """Initialize parameters. 
        Normally should contain ROI, N_distances, etc
        """
        self.N_start = N_start
        self.ROI = ROI
        self.N_files = (N_finish - N_start) 
        self.im_shape = (ROI[3]-ROI[1], ROI[2]-ROI[0])  
        self.Npad = init_Npad(ROI, compression = compNpad)
        
        
    def init_paths(self, data_name, path, N_distances):
        """Generate paths images & flatfields"""
    
        #set data_names
        data_names, ff_names = init_names(data_name, N_distances)
        
        #find images
        imlist = var.im_folder(path)
        
        #set proper paths
        images = np.zeros(N_distances, 'object') 
        flats = np.zeros(N_distances, 'object')
                
        for i in np.arange(len(images)):
            #sort image paths
            images[i] = [path+im for im in imlist if (im.startswith(data_names[i])) and not (im.startswith('.'))]
            flats[i] = [path+im for im in imlist if im.startswith(ff_names[i])]
            
        self.images = images
        self.flats = flats

                        

class F:
    __slots__ = ['shape', 'dtype', 'shared_array_base']
    
    def __init__(self, shape, dtype = 'd'):
        """Create shared value array for processing.
        """
        self.shape = shape
        self.dtype = dtype
        
        ncell = int(np.prod(self.shape))
        self.shared_array_base = Array(dtype, ncell,lock=False)       
        pass
     
        
    




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Nov 21 16:23:15 2019

@author: mpolikarpov
"""
import os
os.environ['OMP_NUM_THREADS'] ='1'
os.environ['OPENBLAS_NUM_THREADS'] = '1'
os.environ['MKL_NUM_THREADS'] = '1'

from maximus48 import var
import numpy as np
from multiprocessing import Array
from maximus48.multiCTF2 import shift_distance as shift


class Processor:
    
    def __init__(self, **kwargs):
        """Initialize parameters. 
        Normally should contain ROI, N_distances, etc
        """
        self.__dict__.update(kwargs)        
        pass
       
        
    def init_parameters(self, **kwargs):
        """Initialize parameters """
        
        self.__dict__.update(kwargs) 
        
# =============================================================================
#         params = dict()
# 
#         for key, value in kwargs.items():
#             params[key] = value
#         return params
# =============================================================================
    
    
    def init_Npad(self, compression = 8):
        """Calculate the Npad for padding
        can be adjusted with compression parameter
        By default, 8 times smaller than ROI
        """
                        
        ROI = self.ROI
        if (ROI[2]-ROI[0])>(ROI[3]-ROI[1]):
            self.Npad = (ROI[2]-ROI[0])//compression       
        else:
            self.Npad = (ROI[3]-ROI[1])//compression   
        
        #return(self.Npad)
        
        
    def init_names(self, data_name, N_distances, first_distance = 1):
        """set proper data_names"""
        
        data_names = []
        ff_names = []
        
        self.N_distances = N_distances
    
        for i in range(first_distance, N_distances + first_distance):
            data_names.append(data_name + '_' + str(i))
            ff_names.append('ff_' + data_name + '_' + str(i))
        
        return data_names, ff_names
    
    
    def init_paths(self, data_name, path, **kwargs):
        """Generate paths images & flatfields"""
        
        #set variables
        param = Processor().init_parameters(**kwargs)
        
        if 'N_distances' in self.__dict__:
            N_distances = self.N_distances
        else:
            N_distances = param['N_distances']
            
        #set data_names
        data_names, ff_names = Processor().init_names(data_name, N_distances)
        
        #find images
        imlist = var.im_folder(path)
        
        #set proper paths
        images = np.zeros(N_distances, 'object') 
        flats = np.zeros(N_distances, 'object')
                
        for i in np.arange(len(images)):
            #sort image paths
            images[i] = [path+im for im in imlist if (im.startswith(data_names[i])) and not (im.startswith('.'))]
            flats[i] = [path+im for im in imlist if im.startswith(ff_names[i])]
            
        self.images = images
        self.flats = flats
        #return images, flats
        

                              
  

class F:
    
    def __init__(self, **kwargs):
        """Initialize parameters. 
        Normally should contain ROI, N_distances, etc
        """
        self.__dict__.update(kwargs)        
        pass
     
             
    def init_shared(self, dtype = 'd', **kwargs):
        '''Create shared value array for processing.'''
        
        param = Processor().init_parameters(**kwargs)
        if 'shape' not in self.__dict__:
            self.shape = param['shape']
            
        ncell = int(np.prod(self.shape))
        self.shared_array_base = Array(dtype, ncell,lock=False)
        self.dtype = dtype
        #return(shared_array_base)
    
    
    def fill_shared(self, data, **kwargs):
        '''fill shared value array with your data'''
        
        param = Processor().init_parameters(**kwargs)
        
        if 'shape' not in self.__dict__:
            try: 
                self.shape = param['shape']
            except:
                print('Error: initialize shape or type it here as shape=desired_value')
        
        if 'shared_array_base' not in self.__dict__:
            print('Error: initialize shared array first')
        
        shared_arr = F().tonumpyarray(shared_array = self.shared_array_base, 
                                      shape = self.shape,
                                      dtype = self.dtype)
        np.copyto(shared_arr, data)
        
        
    
    
    
def tonumpyarray(shared_array, shape, dtype):
    '''Create numpy array from shared memory.'''
    nparray = np.frombuffer(shared_array, dtype=dtype).reshape(shape)    
    #assert nparray.base is shared_array
    return nparray
        
        
    
    
    
    
def correct_shifts(shifts, median_dev = 5):
    """find any bad numbers which are deviate more than 5 pixels from the median
    and correct them to median of the array"""
    
    shifts = tonumpyarray(shifts.shared_array_base, shifts.shape, shifts.dtype)
    
    for i in range(shifts.shape[1]):
        for j in range(shifts.shape[2]):
            shifts[:,i,j] = np.where((abs(shifts[:,i,j] - np.median(shifts[:,i,j])) > 5), np.median(shifts[:,i,j]), shifts[:,i,j])
    print('adjusted shifts') 



def rotaxis(proj, N_steps, ROI_ff = None):
    """calculate the rotation axis comparing 0 and 180 projection shift
    proj: 3D array
    N_steps: projections per degree
    N_rot: number of files that have their mirrors
    ROI_ff: where to make comparison
    
    """
    if not ROI_ff:
        ROI_ff = 0,0, proj.shape[1], proj.shape[0]
        
    cent = []
    N_rot = proj.shape[0] - 180 * N_steps
    
    for i in range(N_rot):
        distances = shift(proj[i, ROI_ff[1]:ROI_ff[3], ROI_ff[0]:ROI_ff[2]], 
              np.flip(proj[i + N_steps*180, ROI_ff[1]:ROI_ff[3], ROI_ff[0]:ROI_ff[2]] ,1))
        cent.append(proj[i].shape[1]/2 + distances[1]/2)
    return cent
    

        




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Aug  7 14:01:47 2018

@author: mpolikarpov, Maxim Polikarpov EMBL HH


compares images with the flatfield with the algorithm described in 
https://www.pyimagesearch.com/2014/09/15/python-compare-two-images/

"""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import numpy as np
from numpy import sqrt, square 
#import time
        
           

    

class SSIM_const(object):
    __slots__ = ['shape', 'mu', 'delta', 'sigma']
    
    """ Calculate flatfield-related constants for the 2D image
    
    Parameters
    __________
    image : ndarray 
        input image data 2D array, 
        can be in (X,Y,N) shape, where X&Y - pixels of one image, N - different images
    mu, delta, sigma: ndarray
        parameters from the article, can be 2D if there were N of images
    
    """
    
    def __init__(self, image):
        """ Constructor """
        
        #self.data = image
        self.shape = image.shape
        self.mu = np.sum(image, axis = (0,1))/(self.shape[0]*self.shape[1])
        self.delta = image-self.mu
        
        razn = square(image - self.mu)
        if len(self.shape) == 2:        
            self.sigma = sqrt(np.sum(razn)/(razn.shape[0]*razn.shape[1]-1))
        
        elif len(self.shape) == 3:
            self.sigma = sqrt(np.sum(razn, axis = (0,1))/(razn.shape[0]*razn.shape[1]-1))
        
        pass


    

class SSIM(object):
    __slots__ = ['data', 'ff']
    """ 
    Calculate SSIM-index for 2D/3D images
    __________
    image : class Image2D 
        can work for 2D and 3D images
       
    out: ndarray
        calculate SSIM indexes for the images
    
    
    
    
    """
    
    def __init__(self, data, flatfield):
        self.data = data
        self.ff = flatfield 
        
        pass
            

    def ssim(self):
        ''' this function calculates SSIM for 2D or 3D case'''
        
        if len(self.ff.shape) == 2:
            #2D case
            C1=0
            C2=0
            
            #claculate sigmaxy first
            k = self.data.delta * self.ff.delta
            k1 = np.sum(k)
            sxy = k1/(k.shape[0]*k.shape[1]-1)
            
            #calculate SSIM
            nominator = (2 * self.data.mu * self.ff.mu + C1) * (2 * sxy + C2)
            denominator1 = (square(self.data.mu) + square(self.ff.mu) + C1) 
            denominator2 = (square(self.data.sigma) + square(self.ff.sigma) + C2)
            
            out = nominator / (denominator1 * denominator2)    
            
                                                
            
        elif len(self.ff.shape) == 3:
            #3D case
            out = np.zeros(self.ff.shape[2])
            
            for i in range(self.ff.shape[2]):
            
                C1=0
                C2=0
                
                #calculate sigmaxy first
                k = self.data.delta * self.ff.delta[:,:,i]
                k1 = np.sum(k)
                sxy = k1/(k.shape[0]*k.shape[1]-1)
                
                #calculate SSIM
                nominator = (2 * self.data.mu * self.ff.mu[i] + C1) * (2 * sxy + C2)
                denominator1 = (square(self.data.mu) + square(self.ff.mu[i]) + C1) 
                denominator2 = (square(self.data.sigma) + square(self.ff.sigma[i]) + C2)
                
                out[i] = nominator / (denominator1 * denominator2)    
        
        return out
            
        
        
        
    
            
           
    
    

    
# =============================================================================
#     def ssim2D(self):
#         #works!
#         if len(self.ff.shape) != 2:
#             raise Exception('Your flatfield.shape is %s but you try to use SSIM function for 2D flatfield-array' %(len(self.ff.shape)))
# 
#         C1=0
#         C2=0
#         
#         #calculate sigmaxy first
#         k = self.data.delta * self.ff.delta
#         k1 = np.sum(k)
#         sxy = k1/(k.shape[0]*k.shape[1]-1)
#         
#         #calculate SSIM
#         nominator = (2 * self.data.mu * self.ff.mu + C1) * (2 * sxy + C2)
#         denominator1 = (square(self.data.mu) + square(self.ff.mu) + C1) 
#         denominator2 = (square(self.data.sigma) + square(self.ff.sigma) + C2)
#         
#         out = nominator / (denominator1 * denominator2)
#         
#         return out  
# =============================================================================
    
    
# =============================================================================
# class SSIM(object):
#     """ Calculates the SSIM and divides the image""" 
#     
#     def __init__(self, data, flatfield):
#         self.data = data
#         self.ff = flatfield
#         pass
#                   
#     
#     def ssim(self):
#         C1=0
#         C2=0
#         
#         #claculate sigmaxy first
#         k = self.data.delta * self.ff.delta
#         k1 = np.sum(k)
#         sxy = k1/(k.shape[0]*k.shape[1]-1)
#         
#         #calculate SSIM
#         nominator = (2 * self.data.mu * self.ff.mu + C1) * (2 * sxy + C2)
#         denominator1 = (square(self.data.mu) + square(self.ff.mu) + C1) 
#         denominator2 = (square(self.data.sigma) + square(self.ff.sigma) + C2)
#         
#         SSIM = nominator / (denominator1 * denominator2)
#         
#         return SSIM   
# =============================================================================
    
    
    
# =============================================================================
# def ssim(ux, uy, sx, sy, deltaX, deltaY):
#     #Calculate similarity index, based on constants#
#     
#     C1=0
#     C2=0
#     
#     #calculate sigmaxy first
#     k = deltaX * deltaY
#     k1 = np.sum(k)
#     sxy = k1/(k.shape[0]*k.shape[1]-1)
#     
#     SSIM = (2*ux*uy+C1)*(2*sxy+C2)/((square(ux)+square(uy)+C1)*(square(sx)+square(sy)+C2))
#     return SSIM;
# =============================================================================

    
    
    
    
    
    
    
    
 



